/* tga3_io.c -- read in 24-bit uncompressed TGA format file, storing
 * data in three arrays.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include "hcon.h"
#define PIXTYPE unsigned char

#define U unsigned int

int tga3_read(Cimg *imp, char *fname)
{
  int ix, iy;
  FILE *infile;
  size_t mem;
  Cimg im;

  unsigned char red, green, blue;
  unsigned char IDLength;             /* length of Identifier String */
  unsigned char CoMaPIXTYPE;            /* 0 = no map */
  unsigned char ImgType;              /* image type (see below for values) */
  unsigned char Index_lo, Index_hi;   /* index of first color map entry */
  unsigned char Length_lo, Length_hi; /* number of entries in color map */
  unsigned char CoSize;               /* size colormap entry (15,16,24,32) */
  unsigned char X_org_lo, X_org_hi;   /* x origin of image */
  unsigned char Y_org_lo, Y_org_hi;   /* y origin of image */
  unsigned char Width_lo, Width_hi;   /* width of image */
  unsigned char Height_lo, Height_hi; /* height of image */
  unsigned char PixelSize;            /* pixel size (8,16,24,32) */
  unsigned char Dbyte;            /* pixel size (8,16,24,32) */
    

    /* ------------- open TGA file ----------- */

  infile = fopen(fname, "rb");
  if (infile == NULL) {
    fprintf(stderr,"Trouble opening TGA file %s\n",fname);
    exit(1);
  }


    /*----------- Read in data from TGA format file ------- */
    /* eighteen-byte header */

    fscanf(infile,"%c",&IDLength);
    fscanf(infile,"%c",&CoMaPIXTYPE);
    fscanf(infile,"%c",&ImgType);
    fscanf(infile,"%c",&Index_lo);
    fscanf(infile,"%c",&Index_hi);
    fscanf(infile,"%c",&Length_lo);
    fscanf(infile,"%c",&Length_hi);
    fscanf(infile,"%c",&CoSize);
    fscanf(infile,"%c",&X_org_lo); fscanf(infile,"%c",&X_org_hi);
    fscanf(infile,"%c",&Y_org_lo); fscanf(infile,"%c",&Y_org_hi);
    fscanf(infile,"%c",&Width_lo);
    fscanf(infile,"%c",&Width_hi);              /* file in lo-byte, hi-byte order b12,b13*/
    fscanf(infile,"%c",&Height_lo);
    fscanf(infile,"%c",&Height_hi);    /* ysize b14, b15 */
    fscanf(infile,"%c",&PixelSize);
    fscanf(infile,"%c",&Dbyte);        /* descriptor byte b17 */
  
  im.xsize = (unsigned int) Width_lo + 256*Width_hi;
  im.ysize = (unsigned int) Height_lo + 256*Height_hi;

  mem = (size_t) sizeof(PIXTYPE) * im.xsize * im.ysize;
  fprintf(stderr,"TGA image type %d PixelSize %d Descriptor %d\n",
    (int)ImgType,PixelSize,Dbyte); 
  fprintf(stderr,"allocating %lu bytes for array\n",(long unsigned) mem);
  im.r = (PIXTYPE *) malloc(mem);
  im.g = (PIXTYPE *) malloc(mem);
  im.b = (PIXTYPE *) malloc(mem);
  if (im.r==NULL || im.g==NULL || im.b==NULL) {
	fprintf(stderr,"Could not allocate memory for %s.\n",fname);
	exit(1);
  }
  for (iy = 0; iy < im.xsize; iy++) {
    for (ix = 0; ix < im.ysize; ix++) {
      if (fscanf(infile,"%c%c%c",&blue,&green,&red) == EOF) {
	fprintf(stderr,"Read error, %s file appears truncated.\n",fname);
	return(1);
      }
	Elr(im,ix,iy) = (PIXTYPE)red;
	Elg(im,ix,iy) = (PIXTYPE)green;
	Elb(im,ix,iy) = (PIXTYPE)blue;
    }
  }
  fclose(infile);
  
  printf("TGA file %s: [%d x %d]\n",fname,im.xsize,im.ysize);
  imp[0] = im;
  return(0);

} /* end tga3_read() */


int tga3_write(Cimg imout, char *fname)
{
  int ix, iy;
  FILE *outfile;
  double temp;

  unsigned char red, green, blue;
  unsigned char IDLength;             /* length of Identifier String */
  unsigned char CoMaPIXTYPE;            /* 0 = no map */
  unsigned char ImgType;              /* image type (see below for values) */
  unsigned char Index_lo, Index_hi;   /* index of first color map entry */
  unsigned char Length_lo, Length_hi; /* number of entries in color map */
  unsigned char CoSize;               /* size colormap entry (15,16,24,32) */
  unsigned char X_org_lo, X_org_hi;   /* x origin of image */
  unsigned char Y_org_lo, Y_org_hi;   /* y origin of image */
  unsigned char Width_lo, Width_hi;   /* width of image */
  unsigned char Height_lo, Height_hi; /* height of image */
  unsigned char PixelSize;            /* pixel size (8,16,24,32) */

    /* ------------- open TGA file ----------- */

  outfile = fopen(fname, "wb");
  if (outfile == NULL) {
    fprintf(stderr,"Trouble opening TGA file %s\n",fname);
    exit(1);
  }

    /*----------- Write out data to TGA format file ------- */
    /* eighteen-byte header */
  
     /* b0-b7 */
    IDLength = 0; CoMaPIXTYPE = 0; ImgType = 2; Index_lo = 0; Index_hi=0;
    Length_lo = 0; Length_hi = 0; CoSize = 0; X_org_lo = 0; X_org_hi = 0;
    temp = imout.xsize / 256;
    Width_hi = (unsigned char) temp;     /* low and high bytes of x width */
    Width_lo = imout.xsize % 256;
    temp = imout.ysize / 256;
    Height_hi = (unsigned char) temp;
    Height_lo = imout.ysize % 256;
    PixelSize = 24;
    Y_org_lo = 0;
    Y_org_hi = 0;

    fprintf(outfile,"%c",IDLength);
    fprintf(outfile,"%c",CoMaPIXTYPE);
    fprintf(outfile,"%c",ImgType);
    fprintf(outfile,"%c",Index_lo);
    fprintf(outfile,"%c", Index_hi);
    fprintf(outfile,"%c",Length_lo);
    fprintf(outfile,"%c",Length_hi);
    fprintf(outfile,"%c",CoSize);
    fprintf(outfile,"%c",X_org_lo); fprintf(outfile,"%c",X_org_hi);
    fprintf(outfile,"%c",Y_org_lo); fprintf(outfile,"%c",Y_org_hi);
    fprintf(outfile,"%c",Width_lo);
    fprintf(outfile,"%c",Width_hi);              /* file in lo-byte, hi-byte order b12,b13*/
    fprintf(outfile,"%c",Height_lo);
    fprintf(outfile,"%c",Height_hi);    /* ysize b14, b15 */
    fprintf(outfile,"%c",PixelSize);
    fprintf(outfile,"%c",(unsigned char) 0x20);        /* descriptor byte b17 */

    for (iy = 0; iy < imout.ysize; iy++) {
     for (ix = 0; ix < imout.xsize; ix++) {
       red =   Elr(imout,ix,iy); 
       green = Elg(imout,ix,iy); 
       blue =  Elb(imout,ix,iy); 
       fprintf(outfile,"%c%c%c",blue,green,red);
     }
    }
  
  fclose(outfile);
  return(0);

} /* end tga3_write() */

